import * as THREE from "three";
import { GLTF, GLTFLoader } from 'three/examples/jsm/loaders/GLTFLoader.js';
import { RGBELoader } from 'three/addons/loaders/RGBELoader.js';

document.addEventListener('DOMContentLoaded',function()
{
	const myScene = new MyScene();
	myScene.run();
});

class MyScene
{
	#scene = new THREE.Scene();
	#renderer = new THREE.WebGLRenderer({ antialias: true, alpha: true });

	#xrManager?: THREE.WebXRManager;
	#reticle?: Reticle;
	#hitTestSource?: XRHitTestSource;

	#balls?: THREE.Group;
	#camera: THREE.PerspectiveCamera;
	#selectEnvMap = <HTMLSelectElement>document.getElementById("envmap");
	#selectMesh = <HTMLSelectElement>document.getElementById("mesh");
	#switchtoxr = <HTMLButtonElement>document.getElementById("switchtoxr");
	#switchfromxr = <HTMLButtonElement>document.getElementById("switchfromxr");

	constructor()
	{
		const viewport = document.getElementById("viewport");
		if (viewport && viewport instanceof HTMLElement)
		{
			this.#camera = new THREE.PerspectiveCamera(70, viewport.offsetWidth / viewport.offsetHeight, 0.01, 20);
			this.#camera.position.set(0, 1.5, 2.2);
			this.#camera.lookAt(0, 1.0, 0);
			this.#scene.add(this.#camera);

			this.#renderer.setPixelRatio(window.devicePixelRatio);
			this.#renderer.outputColorSpace = THREE.LinearSRGBColorSpace;
			this.#renderer.shadowMap.enabled = true;
			this.#renderer.shadowMap.type = THREE.PCFSoftShadowMap;
			this.#renderer.setSize(viewport.offsetWidth, viewport.offsetHeight);
			this.#renderer.setClearColor(0xFFFFFF, 1);
			this.#renderer.autoClear = false;
			viewport.appendChild(this.#renderer.domElement);

			this.#setMesh();
			this.#setEnvMap();
			this.#attachListeners();
		}else throw new Error("Missing viewport element");
	}

	run()
	{
		this.#renderer.setAnimationLoop((timestamp, frame) => 
		{//animation loop to render each frame
			if (this.#xrManager) 
			{
				this.#updateXR(frame);
			}
			this.#renderer.clear(true, true, true);
			this.#renderer.render(this.#scene, this.#camera);
		});
	}

	async #setMesh()
	{
		if (this.#balls)
		{
			this.#scene.remove(this.#balls);
			this.#balls.clear();
			this.#balls = undefined;
		}

		if (this.#selectMesh.value)
		{
			const loader = new GLTFLoader();
			const balls = await loader.loadAsync(this.#selectMesh.value);
			this.#balls = balls.scene;
			this.#balls.visible = typeof this.#xrManager === "undefined";
			this.#scene.add(this.#balls);
		}
	}

	async #attachListeners()
	{
		this.#selectEnvMap.addEventListener("change", this.#setEnvMap.bind(this));
		this.#selectMesh.addEventListener("change", this.#setMesh.bind(this));

		this.#selectEnvMap.classList.remove("displayoff");
		this.#selectMesh.classList.remove("displayoff");

		if (await navigator.xr?.isSessionSupported('immersive-ar'))
		{
			this.#switchtoxr.addEventListener("click", this.#enterXR.bind(this));
			this.#switchfromxr.addEventListener("click", this.#exitXR.bind(this));
			this.#switchtoxr.classList.remove("displayoff");
		}
	}

	async #enterXR()
	{
		this.#scene.remove(this.#camera);
		this.#camera = new THREE.PerspectiveCamera(70, window.innerWidth / window.innerHeight, 0.01, 1000);
		this.#scene.add(this.#camera);

		this.#switchtoxr.classList.add("displayoff");
		this.#renderer.setClearColor(0x000000, 0);

		if (this.#balls) this.#balls.visible = false;
		const session = await navigator.xr?.requestSession( 'immersive-ar', {
			requiredFeatures: ['local', 'hit-test'],
			optionalFeatures: ['dom-overlay'],
			domOverlay: { root: document.body }
		});

		if (session)
		{
			this.#xrManager = this.#renderer.xr;
			this.#xrManager.setReferenceSpaceType("local");
			this.#xrManager.setSession(session);
			this.#reticle = new Reticle(this.#scene);
			this.#xrManager.enabled = true;
			var controller = this.#xrManager.getController(0);
			this.#scene.add(controller);

			const viewerSpace = await session.requestReferenceSpace('viewer');
			if (session && "requestHitTestSource" in session && session.requestHitTestSource)
			{
				this.#hitTestSource = await session.requestHitTestSource({ space: viewerSpace });
				console.log("Hit test source requested and" + (this.#hitTestSource ? " granted" : " not granted"));
			}

			document.body.addEventListener("click", this.#onTap);
		}else throw "Failed to obtain XR session";
		this.#switchfromxr.classList.remove("displayoff");
	}

	#exitXR()
	{
		this.#switchfromxr.classList.add("displayoff");
		document.body.removeEventListener("click", this.#onTap);

		this.#scene.remove(this.#camera);
		const viewport = document.getElementById("viewport") as HTMLElement;
		this.#camera = new THREE.PerspectiveCamera(70, viewport.offsetWidth / viewport.offsetHeight, 0.01, 20);
		this.#camera.position.set(0, 1.5, 2.2);
		this.#camera.lookAt(0, 1.0, 0);
		this.#scene.add(this.#camera);

		if (this.#reticle) 
		{
			this.#reticle.dispose();
			this.#reticle = undefined;
		}
		if (this.#xrManager)
		{
			this.#xrManager.enabled = false;
			const session = this.#xrManager.getSession();

			this.#xrManager.dispose();
			this.#xrManager = undefined;
			this.#hitTestSource?.cancel();
			this.#hitTestSource = undefined;

			if (session)
			{
				session.end();
			}
		}
		this.#camera.position.set(0, 1.5, 2.2);
		this.#camera.lookAt(0, 1.0, 0);
		this.#scene.visible = true;
		if (this.#balls)
		{
			this.#balls.visible = true;
			this.#balls.position.set(0,0,0);
			this.#balls.rotation.set(0,0,0);
		}
		this.#renderer.setClearColor(0xFFFFFF, 1);
		this.#switchtoxr.classList.remove("displayoff");
	}

	#updateXR(frame: XRFrame)
	{
		if (frame && this.#xrManager)
		{
			this.#xrManager.updateCamera(this.#camera);

			const referenceSpace = this.#xrManager.getReferenceSpace();

			if (this.#hitTestSource && this.#reticle && referenceSpace) 
			{
				const hitTestResults = frame.getHitTestResults(this.#hitTestSource);
				if (hitTestResults.length) 
				{
					if (!this.#reticle.visible)
					{
						console.log("Reticle added");
						this.#reticle.visible = true;
					}

					const hit = hitTestResults[0];
					const matrix = hit.getPose(referenceSpace)?.transform.matrix;
					if (matrix)	this.#reticle.setMatrix(Array.from(matrix));
				}else if (this.#reticle.visible)
				{
					this.#reticle.visible = false;
				}
			}
		}
	}

	async #setEnvMap()
	{
		if (this.#scene.environment)
		{
			this.#scene.environment.dispose();
		}
		if (this.#selectEnvMap.value)
		{
			const url = this.#selectEnvMap.value;
			this.#selectEnvMap.disabled = true;

			try
			{
				const loader = /\.hdr$/gi.test(url) ? new RGBELoader() : new THREE.TextureLoader();
				this.#renderer.toneMapping = THREE.NeutralToneMapping;
				const envMapTexture = await loader.loadAsync(url);
				envMapTexture.mapping = THREE.EquirectangularReflectionMapping;
				this.#scene.environment = envMapTexture;
			}catch (err)
			{
				console.error(err);
			}
			this.#selectEnvMap.disabled = false;
		}else
		{
			this.#scene.environment = null;
		}
	}


	#onTap = () =>
	{
		if (this.#reticle?.visible && this.#balls)
		{
			this.#balls.matrixAutoUpdate = true;
			this.#balls.position.setFromMatrixPosition(this.#reticle.matrix);
			this.#balls.visible = true;
		}
	}
}

/**
 * Circle to help the user to place the scene in XR.
 */
class Reticle
{
	#scene?: THREE.Scene;
	#reticle?: THREE.Mesh;

	/**
	 * Creates a reticle mesh with all materials
	 */
	constructor(scene: THREE.Scene)
	{			
		this.#reticle = new THREE.Mesh(
			new THREE.RingGeometry(0.15, 0.2, 32).rotateX(-Math.PI / 2),
			new THREE.MeshBasicMaterial( )
		);
		this.#reticle.name = "reticle";
		this.#reticle.matrixAutoUpdate = false;
		this.#reticle.visible = false;
		this.#scene = scene;
		this.#scene.add(this.#reticle);
	}

	get matrix(): THREE.Matrix4
	{
		if (this.#reticle) return this.#reticle.matrix;
		else throw "Reticle has already been disposed";
	}
	
	setMatrix(matrix: number[])
	{
		if (this.#reticle) this.#reticle.matrix.fromArray(matrix);
	}
	
	get visible()
	{
		if (this.#reticle) return this.#reticle.visible; 
		else throw "Reticle has already been disposed";
	}
	
	set visible(visible: boolean)
	{
		if (this.#reticle) this.#reticle.visible = visible;
	}
	
	/**
	 * Delete material from the scene.
	 */
	dispose()
	{
		if (this.#scene && this.#reticle)
		{
			if (this.#reticle)
			{
				this.#reticle.geometry.dispose();
				if (!Array.isArray(this.#reticle.material)) 
				{
					this.#reticle.material.dispose();
				}else this.#reticle.material.forEach(material => material.dispose());
				this.#scene.remove(this.#reticle);
			}
			this.#reticle = undefined;
			this.#scene = undefined;
		}
	}
}
